//
// Created by nicol on 2018-03-25.
//

#ifdef _WIN32
#include <windows.h>
#include <stdio.h>

#define usleep Sleep
#endif

#include "subscription.h"
#include "connection.h"
#include "credentials.h"
#include "proto_helper.h"
#include "tcp_package.h"
#include "tcp_messages.h"
#include "utils/uuid.h"
#include "utils/buffer.h"
#include "utils/debug.h"
#include "proto.h"

esc_event_appeared_callback_t esc_empty_event_appeared_callback = {0,0};
esc_subscription_dropped_callback_t esc_empty_subscription_dropped_callback = {0,0};

struct st_esc_subscription {
    mutex_t* lock;
    esc_subscription_state_t state;
    esc_connection_t* conn;
    uuid_t* correlation_id;
    esc_event_appeared_callback_t event_appeared;
    esc_subscription_dropped_callback_t subscription_dropped;
    int64_t last_commit_position;
    int64_t last_event_number;
    error_t* error;
    thread_t* worker_thread;
    queue_t* events;
    event_t* stopped;
};

void* esc_subscription_worker_thread(void* arg) {
    esc_subscription_t* subscription = arg;

    bool_t subscribed = BOOL_TRUE;
    while(subscribed) {
        esc_resolved_event_t* ev = queue_dequeue(subscription->events);
        //TODO full logic
        if (ev && subscription->event_appeared.fn) {
            subscription->event_appeared.fn(subscription->event_appeared.ctx, subscription, ev);
        }
        if (ev) esc_resolved_event_destroy(ev);

        mutex_lock(subscription->lock);
        subscribed = subscription->state == ESC_SUBSCRIBED;
        mutex_unlock(subscription->lock);

        usleep(1);
    }

    //TODO reason
    if (subscription->subscription_dropped.fn) {
        subscription->subscription_dropped.fn(subscription->subscription_dropped.ctx, subscription, "", subscription->error);
    }

    //TODO uuid instead of ptr
    printf("Subscription %p worker thread stopped\n", subscription);

    event_set(subscription->stopped);
}

esc_subscription_t* esc_subscription_create(esc_connection_t* conn, uuid_t* correlation_id,
                                            esc_event_appeared_callback_t event_appeared,
                                            esc_subscription_dropped_callback_t subscription_dropped
) {
    esc_subscription_t* subscription = malloc(sizeof(esc_subscription_t));
    subscription->lock = mutex_create();
    subscription->state = ESC_SUBSCRIBING;
    subscription->conn = conn;
    subscription->correlation_id = uuid_copy(correlation_id);
    subscription->event_appeared = event_appeared;
    subscription->subscription_dropped = subscription_dropped;
    subscription->error = 0;
    subscription->worker_thread = thread_create(esc_subscription_worker_thread, subscription);
    subscription->events = queue_create();
    subscription->stopped = event_create();
    return subscription;
}

void esc_subscription_destroy(esc_subscription_t* subscription) {
    esc_subscription_stop(subscription);
    mutex_destroy(subscription->lock);
    uuid_destroy(subscription->correlation_id);
    thread_destroy(subscription->worker_thread);
    queue_destroy(subscription->events);
    event_destroy(subscription->stopped);
    free(subscription);
}

void esc_subscription_handle_tcp_package(esc_subscription_t* subscription, tcp_package_t* recv_pkg) {
    mutex_lock(subscription->lock);

    switch (recv_pkg->command) {
        case MESSAGE_SUBSCRIPTIONCONFIRMATION: {
            EventStore__Client__Messages__SubscriptionConfirmation* recv_msg =
                event_store__client__messages__subscription_confirmation__unpack(&protobuf_allocator, buffer_size(recv_pkg->data), buffer_data(recv_pkg->data));
            subscription->state = ESC_SUBSCRIBED;
            subscription->last_event_number = recv_msg->last_event_number;
            subscription->last_commit_position = recv_msg->last_commit_position;
            event_store__client__messages__subscription_confirmation__free_unpacked(recv_msg, &protobuf_allocator);
            thread_start(subscription->worker_thread);
            break;
        }
        case MESSAGE_STREAMEVENTAPPEARED: {
            esc_resolved_event_t* ev = esc_stream_event_appeared_unpack(recv_pkg->data);
            queue_enqueue(subscription->events, ev);
            break;
        }
        default: {
            subscription->error = error_create(recv_pkg->command, "%s", get_string_for_tcp_message(recv_pkg->command));
            subscription->state = ESC_DROPPED;
            break;
        }
        /*
        case MESSAGE_SUBSCRIPTIONDROPPED: {
            //TODO
            subscription->state = ESC_DROPPED;
            break;
        }
        case MESSAGE_NOTAUTHENTICATED: {
            //TODO
            subscription->state = ESC_DROPPED;
            break;
        }
        case MESSAGE_BADREQUEST: {
            //TODO
            subscription->state = ESC_DROPPED;
            break;
        }
        case MESSAGE_NOTHANDLED: {
            //TODO
            subscription->state = ESC_DROPPED;
            break;
        }*/
    }

    tcp_package_destroy(recv_pkg);

    mutex_unlock(subscription->lock);
}

bool_t esc_find_subscription_by_correlation_id(void* _subscription, void* _correlation_id) {
    esc_subscription_t* subscription = _subscription;
    uuid_t* correlation_id = _correlation_id;

    return (uuid_compare(subscription->correlation_id, correlation_id) == 0) ? BOOL_TRUE : BOOL_FALSE;
}

inline esc_subscription_state_t esc_subscription_state(esc_subscription_t* subscription) {
    return subscription->state;
}

void esc_subscription_stop(esc_subscription_t* subscription) {
    mutex_lock(subscription->lock);
    //TODO drop subscription message instead of direct drop
    //TODO reason
    subscription->state = ESC_DROPPED;
    mutex_unlock(subscription->lock);

    event_wait(subscription->stopped);
}
